﻿'    License: Microsoft Public License (Ms-PL) 
Imports System
Imports System.Collections.Generic
Imports System.Linq
Imports System.Text
Imports IndexFileNameFilter = Lucene.Net.Index.IndexFileNameFilter
Imports Lucene.Net
Imports Lucene.Net.Store
Imports Azure = Microsoft.WindowsAzure.StorageClient
Imports System.Collections.Specialized
Imports System.Diagnostics
Imports System.Runtime.Serialization
Imports System.Runtime.Serialization.Formatters
Imports System.Runtime.Serialization.Formatters.Binary
Imports Microsoft.WindowsAzure
Imports System.Configuration
Imports Microsoft.WindowsAzure.StorageClient
Imports System.Xml.Serialization

Public Class AzureDirectory
    Inherits Directory
    Private _catalog As String
    Private _blobClient As CloudBlobClient
    Private _blobContainer As CloudBlobContainer
    Private _cacheDirectory As Directory

#Region "CTOR"
    Public Sub New(ByVal storageAccount As CloudStorageAccount)
        Me.New(storageAccount, Nothing, Nothing)
    End Sub

    Public Sub New(ByVal storageAccount As CloudStorageAccount, ByVal catalog As String)
        Me.New(storageAccount, catalog, Nothing)
    End Sub

    Public Sub New(ByVal storageAccount As CloudStorageAccount, ByVal catalog As String, ByVal cacheDirectory As Directory)
        If storageAccount Is Nothing Then
            Throw New ArgumentNullException("storageAccount")
        End If

        If String.IsNullOrEmpty(catalog) Then
            _catalog = "lucene"
        Else
            _catalog = catalog.ToLower()
        End If

        _blobClient = storageAccount.CreateCloudBlobClient()
        _initCacheDirectory(cacheDirectory)
    End Sub

    Public ReadOnly Property BlobContainer() As CloudBlobContainer
        Get
            Return _blobContainer
        End Get
    End Property

#If COMPRESSBLOBS Then
		Public Property CompressBlobs() As Boolean
#End If
    Public Sub ClearCache()
        For Each file In _cacheDirectory.List()
            _cacheDirectory.DeleteFile(file)
        Next file
    End Sub

    Public Property CacheDirectory() As Directory
        Get
            Return _cacheDirectory
        End Get
        Set(ByVal value As Directory)
            _cacheDirectory = value
        End Set
    End Property
#End Region

#Region "internal methods"
    Private Sub _initCacheDirectory(ByVal cacheDirectory As Directory)
#If COMPRESSBLOBS Then
			CompressBlobs = True
#End If
        If cacheDirectory IsNot Nothing Then
            ' save it off
            _cacheDirectory = cacheDirectory
        Else
            Dim cachePath As String = System.IO.Path.Combine(Environment.ExpandEnvironmentVariables("%temp%"), "AzureDirectory")
            Dim azureDir As New System.IO.DirectoryInfo(cachePath)
            If Not azureDir.Exists Then
                azureDir.Create()
            End If

            Dim catalogPath As String = System.IO.Path.Combine(cachePath, _catalog)
            Dim catalogDir As New System.IO.DirectoryInfo(catalogPath)
            If Not catalogDir.Exists Then
                catalogDir.Create()
            End If

            _cacheDirectory = FSDirectory.GetDirectory(catalogPath)
        End If

        CreateContainer()
    End Sub

    Public Sub CreateContainer()
        _blobContainer = _blobClient.GetContainerReference(_catalog)

        ' create it if it does not exist
        _blobContainer.CreateIfNotExist()
    End Sub
#End Region

#Region "DIRECTORY METHODS"
    ''' <summary>Returns an array of strings, one for each file in the directory. </summary>
    Public Overrides Function List() As System.String()
        Dim results = From blob In _blobContainer.ListBlobs() _
                    Select blob.Uri.AbsolutePath.Substring(blob.Uri.AbsolutePath.LastIndexOf("/"c) + 1)
        Return results.ToArray()
    End Function

    ''' <summary>Returns true if a file with the given name exists. </summary>
    Public Overrides Function FileExists(ByVal name As System.String) As Boolean
        ' this always comes from the server
        Try
            Dim blob = _blobContainer.GetBlobReference(name)
            blob.FetchAttributes()
            Return True
        Catch
            Return False
        End Try
    End Function

    ''' <summary>Returns the time the named file was last modified. </summary>
    Public Overrides Function FileModified(ByVal name As System.String) As Long
        ' this always has to come from the server
        Try
            Dim blob = _blobContainer.GetBlobReference(name)
            blob.FetchAttributes()
            Return blob.Properties.LastModifiedUtc.ToFileTimeUtc()
        Catch
            Return 0
        End Try
    End Function

    ''' <summary>Set the modified time of an existing file to now. </summary>
    Public Overrides Sub TouchFile(ByVal name As System.String)
        'Dim props As BlobProperties = _blobContainer.GetBlobProperties(name)
        '_blobContainer.UpdateBlobMetadata(props)
        ' I have no idea what the semantics of this should be...hmmmm...
        ' we never seem to get called
        _cacheDirectory.TouchFile(name)
        'SetCachedBlobProperties(props)
    End Sub

    ''' <summary>Removes an existing file in the directory. </summary>
    Public Overrides Sub DeleteFile(ByVal name As System.String)
        Dim blob = _blobContainer.GetBlobReference(name)
        blob.DeleteIfExists()
        Debug.WriteLine(String.Format("DELETE {0}/{1}", _blobContainer.Uri.ToString(), name))

        If _cacheDirectory.FileExists(name & ".blob") Then
            _cacheDirectory.DeleteFile(name & ".blob")
        End If

        If _cacheDirectory.FileExists(name) Then
            _cacheDirectory.DeleteFile(name)
        End If
    End Sub

    ''' <summary>Renames an existing file in the directory.
    ''' If a file already exists with the new name, then it is replaced.
    ''' This replacement should be atomic. 
    ''' </summary>
    Public Overrides Sub RenameFile(ByVal [from] As System.String, ByVal [to] As System.String)
        Try
            Dim blobFrom = _blobContainer.GetBlobReference([from])
            Dim blobTo = _blobContainer.GetBlobReference([to])
            blobTo.CopyFromBlob(blobFrom)
            blobFrom.DeleteIfExists()

            ' we delete and force a redownload, since we can't do this in an atomic way
            If _cacheDirectory.FileExists([from]) Then
                _cacheDirectory.RenameFile([from], [to])
            End If

            ' drop old cached data as it's wrong now
            If _cacheDirectory.FileExists([from] & ".blob") Then
                _cacheDirectory.DeleteFile([from] & ".blob")
            End If
        Catch
        End Try
    End Sub

    ''' <summary>Returns the length of a file in the directory. </summary>
    Public Overrides Function FileLength(ByVal name As System.String) As Long
        Dim blob = _blobContainer.GetBlobReference(name)
        blob.FetchAttributes()

        ' index files may be compressed so the actual length is stored in metatdata
        Dim blobLength As Long
        If Long.TryParse(blob.Metadata("CachedLength"), blobLength) Then
            Return blobLength
        Else
            Return blob.Properties.Length ' fall back to actual blob size
        End If
    End Function

    ''' <summary>Creates a new, empty file in the directory with the given name.
    ''' Returns a stream writing this file. 
    ''' </summary>
    Public Overrides Function CreateOutput(ByVal name As System.String) As IndexOutput
        Dim blob = _blobContainer.GetBlobReference(name)
        Return New AzureIndexOutput(Me, blob)
    End Function

    ''' <summary>Returns a stream reading an existing file. </summary>
    Public Overrides Function OpenInput(ByVal name As System.String) As IndexInput
        Try
            Dim blob As CloudBlob = _blobContainer.GetBlobReference(name)
            blob.FetchAttributes()
            Dim input As New AzureIndexInput(Me, blob)
            Return input
        Catch
            Throw New System.IO.FileNotFoundException(name)
        End Try
    End Function

    ''' <summary>Construct a {@link Lock}.</summary>
    ''' <param name="name">the name of the lock file
    ''' </param>
    Public Overrides Function MakeLock(ByVal name As System.String) As Lock
        Return New AzureLock(name, Me)
    End Function

    Public Overrides Sub ClearLock(ByVal name As String)
        Dim blob = _blobContainer.GetBlobReference(name)
        blob.DeleteIfExists()
        _cacheDirectory.ClearLock(name)
    End Sub

    ''' <summary>Closes the store. </summary>
    Public Overrides Sub Close()
        _blobContainer = Nothing
        _blobClient = Nothing
    End Sub
#End Region

#Region "Azure specific methods"
#If COMPRESSBLOBS Then
		Public Overridable Function ShouldCompressFile(ByVal path As String) As Boolean
			If Not CompressBlobs Then
				Return False
			End If

			Dim ext As String = System.IO.Path.GetExtension(path)
			Select Case ext
				Case ".cfs", ".fdt", ".fdx", ".frq", ".tis", ".tii", ".nrm", ".tvx", ".tvd", ".tvf", ".prx"
					Return True
				Case Else
					Return False
			End Select
		End Function
#End If
    Public Function OpenCachedInputAsStream(ByVal name As String) As StreamInput
        Return New StreamInput(CacheDirectory.OpenInput(name))
    End Function

    Public Function CreateCachedOutputAsStream(ByVal name As String) As StreamOutput
        Return New StreamOutput(CacheDirectory.CreateOutput(name))
    End Function

#End Region
End Class
