﻿'    License: Microsoft Public License (Ms-PL) 
Imports System
Imports System.Collections.Generic
Imports System.Linq
Imports System.Text
Imports System.IO
Imports Lucene.Net
Imports Lucene.Net.Store
Imports Microsoft.WindowsAzure.StorageClient
Imports System.Diagnostics
Imports System.IO.Compression
Imports System.Threading

    Public Class AzureIndexInput
        Inherits IndexInput
        Private _azureDirectory As AzureDirectory
        Private _blobContainer As CloudBlobContainer
        Private _blob As CloudBlob
        Private _name As String

        Private _indexInput As IndexInput
        Private _fileMutex As Mutex

        Public ReadOnly Property CacheDirectory() As Lucene.Net.Store.Directory
            Get
                Return _azureDirectory.CacheDirectory
            End Get
        End Property

        Public Sub New(ByVal azuredirectory As AzureDirectory, ByVal blob As CloudBlob)
            _name = blob.Uri.Segments(blob.Uri.Segments.Length - 1)

#If FULLDEBUG Then
			Debug.WriteLine(String.Format("opening {0} ", _name))
#End If
            _fileMutex = New Mutex(False, _name)
            _fileMutex.WaitOne()
            Try
                _azureDirectory = azuredirectory
                _blobContainer = azuredirectory.BlobContainer
                _blob = blob

                Dim fileName As String = _name

                Dim fFileNeeded As Boolean = False
                If Not CacheDirectory.FileExists(fileName) Then
                    fFileNeeded = True
                Else
                    Dim cachedLength As Long = CacheDirectory.FileLength(fileName)
                    Dim blobLength As Long = blob.Properties.Length
                    Long.TryParse(blob.Metadata("CachedLength"), blobLength)

                    Dim longLastModified As Long = 0
                Dim blobLastModifiedUTC As DateTime = blob.Properties.LastModifiedUtc
                    If Long.TryParse(blob.Metadata("CachedLastModified"), longLastModified) Then
                    blobLastModifiedUTC = New DateTime(longLastModified).ToUniversalTime()
                    End If

                If cachedLength <> blobLength Then
                    fFileNeeded = True
                Else
                    ' there seems to be an error of 1 tick which happens every once in a while 
                    ' for now we will say that if they are within 1 tick of each other and same length 
                    Dim cachedLastModifiedUTC As DateTime = New DateTime(CacheDirectory.FileModified(fileName), DateTimeKind.Local).ToUniversalTime()
                    If cachedLastModifiedUTC <> blobLastModifiedUTC Then
                        Dim timeSpan As TimeSpan = blobLastModifiedUTC.Subtract(cachedLastModifiedUTC)
                        If timeSpan.TotalSeconds > 1 Then
                            fFileNeeded = True
                        Else
#If FULLDEBUG Then
								Debug.WriteLine(timeSpan.TotalSeconds)
#End If
                            ' file not needed
                        End If
                    End If
                End If
                End If

                ' if the file does not exist
                ' or if it exists and it is older then the lastmodified time in the blobproperties (which always comes from the blob storage)
                If fFileNeeded Then
#If COMPRESSBLOBS Then
					If _azureDirectory.ShouldCompressFile(_name) Then
						' then we will get it fresh into local deflatedName 
						' Dim deflatedStream As New StreamOutput(CacheDirectory.CreateOutput(deflatedName))
						Dim deflatedStream As New MemoryStream()

						' get the deflated blob
						_blob.DownloadToStream(deflatedStream)

						Debug.WriteLine(String.Format("GET {0} RETREIVED {1} bytes", _name, deflatedStream.Length))

						' seek back to begininng
						deflatedStream.Seek(0, SeekOrigin.Begin)

						' open output file for uncompressed contents
						Dim fileStream As StreamOutput = _azureDirectory.CreateCachedOutputAsStream(fileName)

						' create decompressor
						Dim decompressor As New DeflateStream(deflatedStream, CompressionMode.Decompress)

						Dim bytes(65534) As Byte
						Dim nRead As Integer = 0
						Do
							nRead = decompressor.Read(bytes, 0, 65535)
							If nRead > 0 Then
								fileStream.Write(bytes, 0, nRead)
							End If
						Loop While nRead = 65535
						decompressor.Close() ' this should close the deflatedFileStream too

						fileStream.Close()

					Else
#End If
                    Dim fileStream As StreamOutput = _azureDirectory.CreateCachedOutputAsStream(fileName)

                    ' get the blob
                    _blob.DownloadToStream(fileStream)

                    fileStream.Flush()
                    Debug.WriteLine(String.Format("GET {0} RETREIVED {1} bytes", _name, fileStream.Length))

                fileStream.Close()

#If COMPRESSBLOBS Then
                End If
#End If
                ' and open it as an input 
                _indexInput = CacheDirectory.OpenInput(fileName)
				Else
#If FULLDEBUG Then
					Debug.WriteLine(String.Format("Using cached file for {0}", _name))
#End If

                ' open the file in read only mode
                _indexInput = CacheDirectory.OpenInput(fileName)
				End If
            Finally
                _fileMutex.ReleaseMutex()
            End Try
        End Sub

        Public Sub New(ByVal cloneInput As AzureIndexInput)
            _fileMutex = New Mutex(False, cloneInput._name)
            _fileMutex.WaitOne()

            Try
#If FULLDEBUG Then
				Debug.WriteLine(String.Format("Creating clone for {0}", cloneInput._name))
#End If
                _azureDirectory = cloneInput._azureDirectory
                _blobContainer = cloneInput._blobContainer
                _blob = cloneInput._blob
                _indexInput = TryCast(cloneInput._indexInput.Clone(), IndexInput)
            Catch e1 As Exception
                ' sometimes we get access denied on the 2nd stream...but not always. I haven't tracked it down yet
                ' but this covers our tail until I do
                Debug.WriteLine(String.Format("Dagnabbit, falling back to memory clone for {0}", cloneInput._name))
            Finally
                _fileMutex.ReleaseMutex()
            End Try
        End Sub

        Public Overrides Function ReadByte() As Byte
            Return _indexInput.ReadByte()
        End Function

        Public Overrides Sub ReadBytes(ByVal b() As Byte, ByVal offset As Integer, ByVal len As Integer)
            _indexInput.ReadBytes(b, offset, len)
        End Sub

        Public Overrides Function GetFilePointer() As Long
            Return _indexInput.GetFilePointer()
        End Function

        Public Overrides Sub Seek(ByVal pos As Long)
            _indexInput.Seek(pos)
        End Sub

        Public Overrides Sub Close()
            _fileMutex.WaitOne()
            Try
#If FULLDEBUG Then
				Debug.WriteLine(String.Format("CLOSED READSTREAM local {0}", _name))
#End If
                _indexInput.Close()
                _indexInput = Nothing
                _azureDirectory = Nothing
                _blobContainer = Nothing
                _blob = Nothing
                GC.SuppressFinalize(Me)
            Finally
                _fileMutex.ReleaseMutex()
            End Try
        End Sub

        Public Overrides Function Length() As Long
            Return _indexInput.Length()
        End Function

        Public Overrides Function Clone() As System.Object
        Dim _clone As IndexInput = Nothing
        Try
            _fileMutex.WaitOne()
            Dim input As New AzureIndexInput(Me)
            _clone = CType(input, IndexInput)
        Catch err As System.Exception
            Debug.WriteLine(err.ToString())
        Finally
            _fileMutex.ReleaseMutex()
        End Try
        Debug.Assert(_clone IsNot Nothing)
        Return _clone
        End Function

    End Class